﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace ElfViewerConsole
{
    class CmdLineArgs
    {
        public enum CMDLINEARGS_RESULT
        {
            CMDLINEARGS_OK,
            CMDLINEARGS_FILENOTFOUND,
            CMDLINEARGS_NONE,
            CMDLINEARGS_INVALID
        };

        public class CmdLineOpt
        {
            public string description;
            public bool active;

            public CmdLineOpt(string desc, bool active)
            {
                this.description = desc;
                this.active = active;
            }
        }

        public string FileName { get; private set; }
        public string InvalidArg { get; private set; }

        public static readonly string OptStrGeneralFileInfo = "-g";
        public static readonly string OptStrELFHeaderShort = "-hs";
        public static readonly string OptStrELFHeaderDetailed = "-hd";
        public static readonly string OptStrProgramHeaderTableShort = "-ps";
        public static readonly string OptStrProgramHeaderTableDetailed = "-pd";
        public static readonly string OptStrSectionHeaderTableShort = "-ss";
        public static readonly string OptStrSectionHeaderTableDetailed = "-sd";
        public static readonly string OptStrSymbolTablesShort = "-ys";
        public static readonly string OptStrSymbolTablesDetailed = "-yd";
        public static readonly string OptStrNoteTables = "-n";

        public static readonly string OptStrAllShort = "-as";
        public static readonly string OptStrAllDetailed = "-ad";

        private Dictionary<string, CmdLineOpt> cmdOptions = new Dictionary<string, CmdLineOpt>()
        {
            { OptStrGeneralFileInfo, new CmdLineOpt("General file information", false) },
            { OptStrELFHeaderShort, new CmdLineOpt("ELF header information (normal)", false) },
            { OptStrELFHeaderDetailed, new CmdLineOpt("ELF header information (detailed)", false) },
            { OptStrProgramHeaderTableShort, new CmdLineOpt("Program header table information (normal)", false) },
            { OptStrProgramHeaderTableDetailed, new CmdLineOpt("Program header table information (detailed)", false) },
            { OptStrSectionHeaderTableShort, new CmdLineOpt("Section header table information (normal)", false) },
            { OptStrSectionHeaderTableDetailed, new CmdLineOpt("Section header table information (detailed)", false) },
            { OptStrSymbolTablesShort, new CmdLineOpt("Symbol header table information (normal)", false) },
            { OptStrSymbolTablesDetailed, new CmdLineOpt("Symbol header table information (detailed)", false) },
            { OptStrNoteTables, new CmdLineOpt("Note table information", false) },

            { OptStrAllShort, new CmdLineOpt("All available information (normal)", false) },
            { OptStrAllDetailed, new CmdLineOpt("All available information (detailed)", false) },
        };

        public CMDLINEARGS_RESULT ProcessArgs(string[] args)
        {
            CMDLINEARGS_RESULT result = CMDLINEARGS_RESULT.CMDLINEARGS_OK;

            if (args.Length == 0)
            {
                result = CMDLINEARGS_RESULT.CMDLINEARGS_NONE;
            }
            else
            {
                string filename = args[0];
                if (!File.Exists(filename))
                {
                    result = CMDLINEARGS_RESULT.CMDLINEARGS_FILENOTFOUND;
                }
                else
                {
                    FileName = filename;
                    if (args.Length == 1)
                    {
                        /* only file specified */
                    }

                    for (int i = 1; i < args.Length; i++)
                    {
                        if (!HandleCmdOption(args[i]))
                        {
                            InvalidArg = args[i];
                            result = CMDLINEARGS_RESULT.CMDLINEARGS_INVALID;
                            break;
                        }
                    }
                }

            }

            return result;
        }

        private bool HandleCmdOption(string option)
        {
            if (OptStrAllShort.Equals(option))
            {
                cmdOptions[OptStrELFHeaderShort].active = true;
                cmdOptions[OptStrProgramHeaderTableShort].active = true;
                cmdOptions[OptStrSectionHeaderTableShort].active = true;
                cmdOptions[OptStrSymbolTablesShort].active = true;
                cmdOptions[OptStrGeneralFileInfo].active = true;
                cmdOptions[OptStrNoteTables].active = true;
                cmdOptions[OptStrAllShort].active = true;

                cmdOptions[OptStrELFHeaderDetailed].active = false;
                cmdOptions[OptStrProgramHeaderTableDetailed].active = false;
                cmdOptions[OptStrSectionHeaderTableDetailed].active = false;
                cmdOptions[OptStrSymbolTablesDetailed].active = false;
                cmdOptions[OptStrAllDetailed].active = false;

                return true;
            }

            if (OptStrAllDetailed.Equals(option))
            {
                cmdOptions[OptStrELFHeaderDetailed].active = true;
                cmdOptions[OptStrProgramHeaderTableDetailed].active = true;
                cmdOptions[OptStrSectionHeaderTableDetailed].active = true;
                cmdOptions[OptStrSymbolTablesDetailed].active = true;
                cmdOptions[OptStrGeneralFileInfo].active = true;
                cmdOptions[OptStrNoteTables].active = true;
                cmdOptions[OptStrAllDetailed].active = true;

                cmdOptions[OptStrELFHeaderShort].active = false;
                cmdOptions[OptStrProgramHeaderTableShort].active = false;
                cmdOptions[OptStrSectionHeaderTableShort].active = false;
                cmdOptions[OptStrSymbolTablesShort].active = false;
                cmdOptions[OptStrAllShort].active = false;

                return true;
            }

            if (cmdOptions.ContainsKey(option))
            {
                cmdOptions[option].active = true;
                return true;
            }
            else
            {
                return false;
            }
        }

        public void PrintUsage()
        {
            Console.Out.WriteLine("Sunshine's ELF Viewer");
            Console.Out.WriteLine("Version: 1.0");
            Console.Out.WriteLine("---------------------");
            Console.Out.WriteLine("Usage: ELFViewerConsole <filename> <options>");
            Console.Out.WriteLine("<Options>:");
            foreach (string key in cmdOptions.Keys)
            {
                Console.Out.WriteLine(" " + key.PadRight(4, ' ') + " : " + cmdOptions[key].description);
            }
            Console.Out.WriteLine();
            Console.Out.WriteLine("Example: ELFViewerConsole test.elf " + OptStrGeneralFileInfo + " " + OptStrProgramHeaderTableDetailed);
        }

        public Dictionary<string, CmdLineOpt> GetOptions()
        {
            return cmdOptions;
        }
    }
}
