﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ElfViewerConsole.ELFModel.ELFFormat.Enums;

namespace ElfViewerConsole.ELFModel.ELFFormat
{
    class ELFElem<T> : IELFElem
    {
        protected long offset;
        protected T value;
        protected string name;
        protected Enum enumValue;
        protected Type enumType;
        protected int? actualLength;

        public ELFElem(long offset, T value, string name, int? actualSizeInBytes)
        {
            this.offset = offset;
            this.value = value;
            this.name = name;
            this.actualLength = actualSizeInBytes;
        }

        public ELFElem(long offset, T value, string name)
            : this(offset, value, name, new Nullable<int>()) { }

        public ELFElem(long offset, T value, string name, Type enumType)
        {
            this.offset = offset;
            this.value = value;
            this.name = name;
            this.enumType = enumType;
            this.enumValue = (Enum)Enum.Parse(enumType, value.ToString());
        }

        public ELFElem(long offset, T value, string name, Type enumType, int? actualSizeInBytes)
        {
            this.offset = offset;
            this.value = value;
            this.name = name;
            this.enumType = enumType;
            this.enumValue = (Enum)Enum.Parse(enumType, value.ToString());
            this.actualLength = actualSizeInBytes;
        }

        public Object ObjectValue
        {
            get { return value; }
        }

        /// <summary>
        /// Actual value of element
        /// </summary>
        public T Value
        {
            get { return value; }
        }

        /// <summary>
        /// File offset of element
        /// </summary>
        public long Offset
        {
            get { return offset; }
        }

        /// <summary>
        /// File offset of element as string
        /// </summary>
        public string OffsetString
        {
            get { return String.Format("{0:X8}", offset); }
        }

        /// <summary>
        /// Name of element
        /// </summary>
        public string Name
        {
            get { return name; }
        }

        /// <summary>
        /// Actual value of element as string
        /// </summary>
        public string ValueString()
        {
            if (this.actualLength.HasValue)
            {
                string formatString = "{0:X" + this.actualLength.Value*2 + "}";
                return String.Format(formatString, value);
            }
            else
            {
                if (value is byte)
                {
                    return String.Format("{0:X2}", value);
                }
                else if (value is ushort)
                {
                    return String.Format("{0:X4}", value);
                }
                else if (value is uint)
                {
                    return String.Format("{0:X8}", value);
                }
                else if (value is string)
                {
                    return String.Format("{0}", value);
                }
                else if (value is ulong)
                {
                    return String.Format("{0:X16}", value);
                }
            }
            return base.ToString();
        }

        /// <summary>
        /// Size of element in bytes
        /// </summary>
        /// <returns></returns>
        public int SizeInBytes()
        {
            if (value is byte)
                return 1;
            else if (value is ushort)
                return 2;
            else if (value is uint)
                return 4;
            else if (value is ulong)
                return 8;
            else return -1;
        }

        /// <summary>
        /// Value of element where each byte is interpreted as Ascii value
        /// </summary>
        /// <returns></returns>
        public string NumValAsAscii()
        {
            if (value is byte)
            {
                byte val = Convert.ToByte(value);
                return Char.ToString((char)val);
            }
            else if (value is ushort)
            {
                ushort val = Convert.ToUInt16(value);
                return Char.ToString((char)(val & 0x00FF)) + Char.ToString((char)(val >> 8));
            }
            else if (value is uint)
            {
                uint val = Convert.ToUInt32(value);
                return Char.ToString((char)(val & 0x000000FF)) +
                       Char.ToString((char)((val & 0x0000FF00) >> 8)) +
                       Char.ToString((char)((val & 0x00FF0000) >> 16)) +
                       Char.ToString((char)((val & 0xFF000000) >> 24));
            }
            else
            {
                throw new NotImplementedException();
            }
        }

        public bool IsEnumValue()
        {
            return this.enumValue != null;
        }

        public Enum GetEnum()
        {
            return this.enumValue;
        }

        public Type GetEnumType()
        {
            return this.enumType;
        }

        public bool HasEnumFlagsAttibrute()
        {
            return (IsEnumValue() && (GetEnumType().GetCustomAttributes(typeof(FlagsAttribute), false).Length > 0));
        }

        public string GetEnumDescription()
        {
            if (GetEnumType().GetCustomAttributes(typeof(FlagsAttribute), false).Length > 0)
            {
                // enum has the Flags attribute, the description of one particular enum value has no sense

                if (GetEnumType().Equals(typeof(EELFShFlags)))
                {
                    return ((EELFShFlags)GetEnum()).ToFriendlyString();
                }
                else if (GetEnumType().Equals(typeof(EELFPFlags)))
                {
                    return ((EELFPFlags)GetEnum()).ToFriendlyString();
                }

                return "";
            }
            else
            {
                if (Enum.IsDefined(GetEnumType(), GetEnum()))
                {
                    return GetEnum().ToDescriptionString();
                }
                else
                {
                    return "Unknown";
                }
            }
        }

    }
}
