﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ELFViewerConsole.ELFModel;
using ElfViewerConsole.ELFModel.ELFFormat.Enums;

namespace ElfViewerConsole.ELFModel.ELFFormat
{
    class ELFHeader : IDisposable
    {
        /// <summary>
        /// File identification
        /// </summary>
        public ELFElem32 E_ident_mag { get; private set; }
        /// <summary>
        /// File class
        /// </summary>
        public ELFElem8Enum<EELFIdentClass> E_ident_class { get; private set; }
        /// <summary>
        /// Data encoding
        /// </summary>
        public ELFElem8Enum<EELFIdentData> E_ident_data { get; private set; }
        /// <summary>
        /// Version number
        /// </summary>
        public ELFElem8Enum<EELFVersion> E_ident_version { get; private set; }
        /// <summary>
        /// Operating system
        /// </summary>
        public ELFElem8Enum<EELFOsAbi> E_ident_osabi { get; private set; }
        /// <summary>
        /// Application binary version
        /// </summary>
        public ELFElem8 E_ident_osabiver { get; private set; }
        /// <summary>
        /// Padding
        /// </summary>
        public ELFElemByteArray E_ident_pad { get; private set; }

        /* offset 0x10 */

        /// <summary>
        /// Object file type
        /// </summary>
        public ELFElem16Enum<EELFType> E_type { get; private set; }
        /// <summary>
        /// Required architecture
        /// </summary>
        public ELFElem16Enum<EELFMachine> E_machine { get; private set; }
        /// <summary>
        /// Object file version
        /// </summary>
        public ELFElem32Enum<EELFVersion> E_version { get; private set; }
        /// <summary>
        /// Virtual address of entry point
        /// </summary>
        public ELFElem64 E_Entry { get; private set; }
        /// <summary>
        /// File offset of program header table
        /// </summary>
        public ELFElem64 E_Phoff { get; private set; }

        /* offset 0x20 */

        /// <summary>
        /// File offset of section header table
        /// </summary>
        public ELFElem64 E_Shoff { get; private set; }
        /// <summary>
        /// Processor-specific flags 
        /// </summary>
        public ELFElem32 E_Flags { get; private set; }
        /// <summary>
        /// Size of ELF header in bytes
        /// </summary>
        public ELFElem16 E_Ehsize { get; private set; }
        /// <summary>
        /// Size of one entry in program header table in bytes
        /// </summary>
        public ELFElem16 E_PhEntSize { get; private set; }
        /// <summary>
        /// Number of entries in the program header table
        /// </summary>
        public ELFElem16 E_PhNum { get; private set; }
        /// <summary>
        /// Size of one entry in section header table in bytes
        /// </summary>
        public ELFElem16 E_ShEntSize { get; private set; }
        /// <summary>
        /// Number of entries in the section header table
        /// </summary>
        public ELFElem16 E_ShNum { get; private set; }
        /// <summary>
        /// Section header table index that is associated with the section name string table
        /// </summary>
        public ELFElem16 E_ShStrndx { get; private set; }

        private ELFFile ElfFile;

        public ELFHeader( ELFFile elfFile)
        {
            this.ElfFile = elfFile;
        }

        internal void Read(IFileAccess fDataReader)
        {
            fDataReader.Position = 0;

            ELFElem8 E_ident_mag0 = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(fDataReader.Position), "EI_MAG0");
            ELFElem8 E_ident_mag1 = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(fDataReader.Position+1), "EI_MAG1");
            ELFElem8 E_ident_mag2 = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(fDataReader.Position+2), "EI_MAG2");
            ELFElem8 E_ident_mag3 = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(fDataReader.Position+3), "EI_MAG3");
            E_ident_mag = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "EI_MAG");
            if ((E_ident_mag0.Value != 0x7F) || (E_ident_mag1.Value != 0x45) || (E_ident_mag2.Value != 0x4C) || (E_ident_mag3.Value != 0x46))
            {
                ElfFile.Valid = false;
                return;
            }

            E_ident_class = new ELFElem8Enum<EELFIdentClass>(fDataReader.Position, fDataReader.ReadByte(), "EI_CLASS");

            E_ident_data = new ELFElem8Enum<EELFIdentData>(fDataReader.Position, fDataReader.ReadByte(), "EI_DATA");
            if ((EELFIdentData)E_ident_data.GetEnum() == EELFIdentData.DATA2MSB)
            {
                fDataReader.SetEndianess(IFileAccessEndianess.BIG_ENDIAN);
            }
            else
            {
                fDataReader.SetEndianess(IFileAccessEndianess.LITTLE_ENDIAN);
            }
            E_ident_version = new ELFElem8Enum<EELFVersion>(fDataReader.Position, fDataReader.ReadByte(), "EI_VERSION");
            E_ident_osabi = new ELFElem8Enum<EELFOsAbi>(fDataReader.Position, fDataReader.ReadByte(), "EI_OSABI");
            E_ident_osabiver = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(), "EI_OSABIVER");
            E_ident_pad = new ELFElemByteArray(fDataReader.Position, fDataReader.ReadByteArray(7), 7, "EI_PAD");

            E_type = new ELFElem16Enum<EELFType>(fDataReader.Position, fDataReader.ReadUInt16(), "E_TYPE");
            E_machine = new ELFElem16Enum<EELFMachine>(fDataReader.Position, fDataReader.ReadUInt16(), "E_MACHINE");
            E_version = new ELFElem32Enum<EELFVersion>(fDataReader.Position, fDataReader.ReadUInt32(), "E_VERSION");

            if (ElfFile.Is64Bit())
            {
                E_Entry = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "E_ENTRY");
                E_Phoff = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "E_PHOFF");
                E_Shoff = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "E_SHOFF");
            }
            else
            {
                E_Entry = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "E_ENTRY", 4);
                E_Phoff = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "E_PHOFF", 4);
                E_Shoff = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "E_SHOFF", 4);
            }
            E_Flags = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "E_FLAGS");
            E_Ehsize = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "E_EHSIZE");
            E_PhEntSize = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "E_PHENTSIZE");
            E_PhNum = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "E_PHNUM");
            E_ShEntSize = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "E_SHENTSIZE");
            E_ShNum = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "E_SHNUM");
            E_ShStrndx = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "E_SHSTRNDX");
        }

        public IEnumerator<IELFElem> GetEnumerator()
        {
            yield return E_ident_mag;
            yield return E_ident_class;
            yield return E_ident_data;
            yield return E_ident_version;
            yield return E_ident_osabi;
            yield return E_ident_osabiver;
            yield return E_ident_pad;

            yield return E_type;
            yield return E_machine;
            yield return E_version;
            yield return E_Entry;
            yield return E_Phoff;
            yield return E_Shoff;
            yield return E_Flags;
            yield return E_Ehsize;
            yield return E_PhEntSize;
            yield return E_PhNum;
            yield return E_ShEntSize;
            yield return E_ShNum;
            yield return E_ShStrndx;
        }

        public void Dispose()
        {
            /* nothing to do - IDisposable must be implemented for using the 'using' keyword */
        }
    }
}
