﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ELFViewerConsole.ELFModel;
using ElfViewerConsole.ELFModel.ELFFormat.Enums;
using System.Collections;

namespace ElfViewerConsole.ELFModel.ELFFormat
{
    class ELFSectionHeaderTable : IDisposable, IEnumerable
    {
        private ELFFile ElfFile;

        public int HeaderTableIndex { get; private set; }

        /// <summary>
        /// Index into string table
        /// </summary>
        public ELFElem32 Sh_name { get; private set; }
        /// <summary>
        /// Section Type
        /// </summary>
        public ELFElem32Enum<EELFShType> Sh_type { get; private set; }
        /// <summary>
        /// Section Flag
        /// </summary>
        public ELFElem64Enum<EELFShFlags> Sh_flags { get; private set; }
        /// <summary>
        /// Address of section in image memory
        /// </summary>
        public ELFElem64 Sh_Addr { get; private set; }
        /// <summary>
        /// Byte offset in file of start of section
        /// </summary>
        public ELFElem64 Sh_Offset { get; private set; }
        /// <summary>
        /// Section size in bytes
        /// </summary>
        public ELFElem64 Sh_Size { get; private set; }
        /// <summary>
        /// Link to another section
        /// </summary>
        public ELFElem32 Sh_Link { get; private set; }
        /// <summary>
        /// Additional info, related to Sh_Link
        /// </summary>
        public ELFElem32 Sh_Info { get; private set; }
        /// <summary>
        /// Address alignment constraint
        /// </summary>
        public ELFElem64 Sh_Addralign { get; private set; }
        /// <summary>
        /// Size of byte of fixed-member entry
        /// </summary>
        public ELFElem64 Sh_Entsize { get; private set; }

        // reference to note section, null if it does not exist
        public ELFNoteTable NoteSection { get; set; }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="index"></param>
        /// <param name="elfFile"></param>
        public ELFSectionHeaderTable(int index, ELFFile elfFile)
        {
            this.ElfFile = elfFile;
            this.HeaderTableIndex = index;
        }

        public void Dispose()
        {
            /* nothing to do */
        }

        internal void Read()
        {
            uint filePos = 0;
            IFileAccess fDataReader = ElfFile.FileData;

            try
            {
                filePos = fDataReader.Position;

                /* move to start of program header table */
                fDataReader.Position = (uint)(ElfFile.ElfFileHeader.E_Shoff.Value + (ulong)(HeaderTableIndex * ElfFile.ElfFileHeader.E_ShEntSize.Value));
                if (ElfFile.Is64Bit())
                {
                    Sh_name = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "SH_NAME");
                    Sh_type = new ELFElem32Enum<EELFShType>(fDataReader.Position, fDataReader.ReadUInt32(), "SH_TYPE");
                    Sh_flags = new ELFElem64Enum<EELFShFlags>(fDataReader.Position, fDataReader.ReadUInt64(), "SH_FLAGS");
                    Sh_Addr = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "SH_ADDR");
                    Sh_Offset = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "SH_OFFSET");
                    Sh_Size = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "SH_SIZE");
                    Sh_Link = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "SH_LINK");
                    Sh_Info = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "SH_INFO");
                    Sh_Addralign = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "SH_ADDRALIGN");
                    Sh_Entsize = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "SH_ENTSIZE");
                }
                else
                {
                    Sh_name = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "SH_NAME");
                    Sh_type = new ELFElem32Enum<EELFShType>(fDataReader.Position, fDataReader.ReadUInt32(), "SH_TYPE");
                    Sh_flags = new ELFElem64Enum<EELFShFlags>(fDataReader.Position, fDataReader.ReadUInt32(), "SH_FLAGS", 4);
                    Sh_Addr = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "SH_ADDR", 4);
                    Sh_Offset = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "SH_OFFSET", 4);
                    Sh_Size = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "SH_SIZE", 4);
                    Sh_Link = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "SH_LINK");
                    Sh_Info = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "SH_INFO");
                    Sh_Addralign = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "SH_ADDRALIGN", 4);
                    Sh_Entsize = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "SH_ENTSIZE", 4);
                }

            }
            finally
            {
                fDataReader.Position = filePos;
            }
        }


        public IEnumerator GetEnumerator()
        {
            yield return Sh_name;
            yield return Sh_type;
            yield return Sh_flags;
            yield return Sh_Addr;
            yield return Sh_Offset;
            yield return Sh_Size;
            yield return Sh_Link;
            yield return Sh_Info;
            yield return Sh_Addralign;
            yield return Sh_Entsize;
        }

        public string GetSectionName()
        {
            if (ElfFile.ExistsSectionNameStringTable())
            {
                int strSecTblIdx = ElfFile.GetSectionNameStringTableIndex();

                ELFSectionHeaderTable strSection = ElfFile.SectionHeaderTables[strSecTblIdx];

                uint filePosToSecName = (uint)(strSection.Sh_Offset.Value + Sh_name.Value);
                string sectionName = ElfFile.FileData.ReadAscii(filePosToSecName, (int)(strSection.Sh_Offset.Value + strSection.Sh_Size.Value - filePosToSecName));

                return sectionName;
            }
            else
            {
                return "-";
            }
        }

        public bool IsNoteSection()
        {
            return NoteSection != null;
        }
    }
}
