﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ELFViewerConsole.ELFModel;
using System.Collections;
using ElfViewerConsole.ELFModel.ELFFormat.Enums;

namespace ElfViewerConsole.ELFModel.ELFFormat
{
    class ELFSymbolTableEntry : IEnumerable
    {
        private ELFFile ElfFile;
        private ELFSymbolTable ParentSymbolTable;

        public int EntryIndex { get; private set; }

        /// <summary>
        /// File offset into symbol string table, which holds the character representations of the symbol names
        /// </summary>
        public ELFElem32 St_name { get; private set; }

        /// <summary>
        /// The value of the associated symbol, depends on context
        /// </summary>
        public ELFElem64 St_value { get; private set; }

        /// <summary>
        /// Size of object
        /// </summary>
        public ELFElem64 St_size { get; private set; }

        /// <summary>
        /// The symbol's type and binding attributes
        /// </summary>
        public ELFElem8  St_info { get; private set; }

        /// <summary>
        /// Visibility of symbol.
        /// </summary>
        public ELFElem8  St_other { get; private set; }

        /// <summary>
        /// Linked section header table index
        /// </summary>
        public ELFElem16 St_shndx { get; private set; }

        public ELFSymbolTableEntry(int index, ELFFile elfFile, ELFSymbolTable symTable)
        {
            this.ElfFile = elfFile;
            this.EntryIndex = index;
            this.ParentSymbolTable = symTable;
        }

        internal void Read(IFileAccess fDataReader, uint startPos)
        {
            uint filePos = 0;

            try
            {
                filePos = fDataReader.Position;
                fDataReader.Position = startPos;
                if (ElfFile.Is64Bit())
                {
                    St_name = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "ST_NAME");
                    St_info = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(), "ST_INFO");
                    St_other = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(), "St_OTHER");
                    St_shndx = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "ST_SHNDX");
                    St_value = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "ST_VALUE");
                    St_size = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt64(), "ST_SIZE");
                }
                else
                {
                    St_name = new ELFElem32(fDataReader.Position, fDataReader.ReadUInt32(), "ST_NAME");
                    St_value = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "ST_VALUE", 4);
                    St_size = new ELFElem64(fDataReader.Position, fDataReader.ReadUInt32(), "ST_SIZE", 4);
                    St_info = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(), "ST_INFO");
                    St_other = new ELFElem8(fDataReader.Position, fDataReader.ReadByte(), "St_OTHER");
                    St_shndx = new ELFElem16(fDataReader.Position, fDataReader.ReadUInt16(), "ST_SHNDX");
                }
            }
            finally
            {
                fDataReader.Position = filePos;
            }
        }

        public IEnumerator GetEnumerator()
        {
            yield return St_name;
            yield return St_value;
            yield return St_size;
            yield return St_info;
            yield return St_other;
            yield return St_shndx;
        }

        public string GetSymbolName()
        {
            if (St_name.Value != 0)
            {
                uint symStringTableIndex = ParentSymbolTable.SectionHeaderTable.Sh_Link.Value;
                if (symStringTableIndex < ElfFile.GetNumOfSections())
                {
                    // file offset of beginning of section of symbol names
                    uint symStringTableFileOffset = (uint)ElfFile.SectionHeaderTables[(int)(symStringTableIndex)].Sh_Offset.Value;
                    // file offset to beginning of symbol name string
                    uint secNameStringFileOffset = symStringTableFileOffset + St_name.Value;

                    string sectionName = ElfFile.FileData.ReadAscii(secNameStringFileOffset,
                        (int)(symStringTableFileOffset + ElfFile.SectionHeaderTables[(int)(symStringTableIndex)].Sh_Size.Value - secNameStringFileOffset));

                    return sectionName;
                }
                else
                {
                    return "-";
                }
            }
            else
            {
                return "-";
            }
        }


        public EELFSymTabBind GetBindingInfo()
        {
            EELFSymTabBind binding = (EELFSymTabBind)(St_info.Value >> 4);
            return binding;
        }

        public string GetBindingInfoAsString()
        {
            return GetBindingInfo().ToDescriptionString();
        }

        public EELFSymTabType GetTypeInfo()
        {
            EELFSymTabType typeInfo = (EELFSymTabType)(St_info.Value & 0x0F);
            return typeInfo;
        }

        public string GetTypeInfoAsString()
        {
            return GetTypeInfo().ToDescriptionString();
        }

        public EELFSymTabVis GetVisibilityInfo()
        {
            EELFSymTabVis typeInfo = (EELFSymTabVis)(St_other.Value & 0x03);
            return typeInfo;
        }

        public string GetVisibilityInfoAsString()
        {
            return GetVisibilityInfo().ToDescriptionString();
        }

    }
}
